import datetime
import glob
import json
import os
import pathlib
import re
from typing import Any, Dict, List, Optional

from appdirs import AppDirs  # type: ignore
import requests
import requests_cache  # type: ignore
from tarsafe import TarSafe  # type: ignore

from ochrona.log import OchronaLogger

# Cache settings
expire_after = datetime.timedelta(hours=1)
requests_cache.install_cache("db_cache", expire_after=expire_after)

RELEASES_URL = (
    "https://api.github.com/repos/ochronasec/ochrona_python_vulnerabilities/releases"
)
VULN_PATTERN = "\.\/vulns\/({})\-[A-Z\-0-9]+\.json"


class VulnDB:

    latest_version: Optional[str] = None
    latest_db_path: Optional[str] = None
    _user_app_dir: str = AppDirs("Ochrona", "Ochrona").user_data_dir

    def __init__(self, logger: OchronaLogger):
        self._logger = logger
        self._create_user_app_dir()
        self._check_local_db_present()
        if self.latest_version is not None:
            if self._is_update_available():
                self._logger.debug(f"More recent version of DB found, will update")
                self._update_db()
        else:
            self._download_latest_db()

    def _is_update_available(self):
        try:
            releases = requests.get(RELEASES_URL).json()
            top_version = max([r.get("name") for r in releases])
            return top_version > self.latest_version
        except Exception as ex:
            self._logger.error(f"Error fetching new releases: {ex}")
            return False

    def _check_local_db_present(self):
        files = glob.glob(f"{self.user_app_dir}/*tar.gz")
        if len(files) > 0:
            version = pathlib.Path(files[0]).name.replace(".tar.gz", "")
            self.latest_version = version
            self.latest_db_path = f"{self.user_app_dir}/{self.latest_version}.tar.gz"
            self._logger.debug(f"DB instance found: {version}")

    def _download_latest_db(self):
        releases = requests.get(RELEASES_URL).json()
        sorted_releases = sorted(releases, key=lambda r: r["name"], reverse=True)
        r = requests.get(
            sorted_releases[0].get("assets")[0].get("browser_download_url")
        )
        with open(
            f"{self.user_app_dir}/{sorted_releases[0].get('assets')[0].get('name')}",
            "wb",
        ) as f:
            f.write(r.content)
        self.latest_version = (
            sorted_releases[0].get("assets")[0].get("name").replace(".tar.gz", "")
        )
        self.latest_db_path = f"{self.user_app_dir}/{self.latest_version}.tar.gz"
        self._logger.debug(f"DB upgraded to {self.latest_version}")

    def _create_user_app_dir(self):
        os.makedirs(self.user_app_dir, exist_ok=True)

    def _update_db(self):
        self._delete_old_dbs()
        self._download_latest_db()

    def _delete_old_dbs(self):
        files = glob.glob(f"{self.user_app_dir}/*tar.gz")
        for file_ in files:
            os.remove(file_)

    def lookup_by_name(self, name: str) -> List[Dict[str, Any]]:
        potential_vuln_paths = []
        potential_vulns = []
        with TarSafe.open(self.latest_db_path, "r:gz") as tar:
            vulns = tar.getmembers()
            for vuln in vulns:
                if re.match(VULN_PATTERN.format(name), vuln.name):
                    potential_vuln_paths.append(vuln.name)

            for vuln_path in potential_vuln_paths:
                potential_vulns.append(json.loads(tar.extractfile(vuln_path).read()))
        self._logger.debug(
            f"Found {len(potential_vulns)} vulnerabilities potentially affecting package: {name}"
        )
        return potential_vulns

    @property
    def user_app_dir(self):
        return self._user_app_dir



import ochrona.eval.policy.package_name
import ochrona.eval.policy.license_type

from ochrona.const import PYTHON_LICENSE_TYPE_POLICY
from ochrona.const import PYTHON_PACKAGE_NAME_POLICY


def policy_evaluate(dependencies, policies):
    """
    Evauluates dependencies for policy violations.
    """
    violations = []
    for policy in policies:
        policy_type = policy.get("policy_type")
        if policy_type == PYTHON_PACKAGE_NAME_POLICY:
            violations += package_name.evaluate(
                dependency_list=dependencies.flat_list, policy=policy
            )
        elif policy_type == PYTHON_LICENSE_TYPE_POLICY:
            violations += license_type.evaluate(
                dependency_list=dependencies.dependencies, policy=policy
            )
    return violations


POLICY_SCHEMAS = {
    PYTHON_PACKAGE_NAME_POLICY: package_name.SCHEMA,  # type: ignore
    PYTHON_LICENSE_TYPE_POLICY: license_type.SCHEMA,  # type: ignore
}



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

import json
import os
import subprocess
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

from ochrona.client import pypi_fetch
from ochrona.const import INVALID_SPECIFIERS, EQUALS_SPECIFIER
from ochrona.eval.eval import resolve
from ochrona.exceptions import OchronaImportException
from ochrona.model.confirmed_vulnerability import Vulnerability
from ochrona.model.dependency_set import DependencySet
from ochrona.parser import Parsers
from ochrona.log import OchronaLogger


class SafeImport:
    def __init__(self, logger: OchronaLogger):
        self._logger = logger

    def install(self, package: str):
        """
        Checks a package and if it's safe, imports via pip.

        :param package: a package string with an optional version specified
        :return: bool
        """
        if os.path.isfile(package):
            parsers = Parsers()
            packages = parsers.requirements.parse(package)
            if self._check_package(packages):
                self._logger.info(
                    f"No vulnerabilities found for {', '.join(packages)}, install proceeding."
                )
                self._install_file(package)
            else:
                self._logger.error(
                    f"Import of {', '.join(packages)} aborted due to detected vulnerabilities."
                )
        else:
            if self._check_package(package):
                self._logger.info(
                    f"No vulnerabilities found for {package}, install proceeding."
                )
                self._install(package=package)
            else:
                self._logger.error(
                    f"Import of {package} aborted due to detected vulnerabilities."
                )

    def _check_package(self, package: Union[str, list]) -> bool:
        """
        Calls the analysis API endpoint with a package to see if a package is safe.

        :param package: a package string with an optional version specified
        :return: bool
        """
        vuln_results: DependencySet
        if isinstance(package, str):
            if any(spec in package for spec in INVALID_SPECIFIERS):
                raise OchronaImportException(
                    f"An invalid specifier was found in {package}, either specify the package without a version or pin to a single version using `name==version`."
                )
            parts = package.split(EQUALS_SPECIFIER)
            if len(parts) == 1:
                package = self._get_most_recent_version(package=package)
            vuln_results = resolve(dependencies=[package], logger=self._logger)
        elif isinstance(package, list):
            vuln_results = resolve(dependencies=package, logger=self._logger)
        if len(vuln_results.confirmed_vulnerabilities) > 0:
            self._logger.info(
                f"A full list of packages that would be installed, included dependencies: {', '.join(vuln_results.flat_list)}"
            )
            self._logger.error(
                f"""Vulerabilities found related to import:\n{''.join([self._format_vulnerability(v) for v in vuln_results.confirmed_vulnerabilities])}"""
            )
            return False
        self._logger.info(
            f"A full list of packages to be installed, included dependencies: {', '.join(vuln_results.flat_list)}"
        )
        return True

    def _get_most_recent_version(self, package: str) -> str:
        """
        If a package does not have a version specified we will assume that the latest version
        should be used.

        :param package: a package string without a version
        :return: str
        """
        fetched = pypi_fetch(package=package)
        latest_version = (
            fetched.get("info", {}).get("version", "") if fetched is not None else ""
        )
        if latest_version != "":
            return f"{package}{EQUALS_SPECIFIER}{latest_version}"
        else:
            self._logger.warn("Unable to reach Pypi to confirm latest version")
            return package

    def _install(self, package: str) -> bool:
        """
        Call pip to install a package

        :param package: a package string with a version
        :return: bool
        """
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
            return True
        except subprocess.CalledProcessError as ex:
            raise OchronaImportException("Error during pip install") from ex

    def _install_file(self, file_path: str) -> bool:
        """
        Call pip to install a requirements.txt style

        :param file_path: path to requirements file
        :return: bool
        """
        try:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "-r", file_path]
            )
            return True
        except subprocess.CalledProcessError as ex:
            raise OchronaImportException("Error during pip install") from ex

    def _format_vulnerability(self, vulnerability: Vulnerability) -> str:
        """
        Formats a vulnerability to a friendly output

        :param vulnerability: vulnerability results
        :return: str
        """
        return f"""\nVulnerability Detected on package to be installed!
    Package: {vulnerability.name}
    ID: {vulnerability.cve_id}
    Description: {vulnerability.description}
    Ochrona Severity: {vulnerability.ochrona_severity_score}
                """



from datetime import timedelta
from typing import Any, Dict, Optional

import requests
import requests_cache  # type: ignore

# Cache settings
expire_after = timedelta(hours=1)
requests_cache.install_cache("pypi_cache", expire_after=expire_after)


def pypi_fetch(package: str) -> Optional[Dict[str, Any]]:
    """
    Get call to Pypi to fetch package detafgils
    """
    url = "https://pypi.org/pypi/{}/json".format(package)
    response = requests.request("GET", url)
    if response.status_code == 200:
        json = response.json()
        if json:
            return json
    return None



import collections
from typing import Any, Dict, List

from ochrona.model.confirmed_vulnerability import Vulnerability
from ochrona.model.dependency import Dependency


class DependencySet:
    """
    A Top level dependencies object for a scan.
    """

    _dependencies: List[Dependency] = []
    _flat_list: List[str] = []
    _confirmed_vulnerabilities: List[Vulnerability] = []
    _policy_violations: List[Dict[str, Any]] = []

    def __init__(self, dependencies=[]):
        self._dependencies = dependencies
        self._flat_list = self._flatten_dependencies(dependencies)

    def _flatten_dependencies(self, dependencies):
        """
        Flattens resolved transitive dependencies into a flat list.
        """
        unique_packages = {}
        for dependency in dependencies:
            if dependency.name not in unique_packages:
                unique_packages[dependency.name] = dependency
            else:
                if dependency.version > unique_packages[dependency.name].version:
                    unique_packages[dependency.name] = dependency
        return list([k.full for k in unique_packages.values()])

    @property
    def dependencies(self):
        return self._dependencies

    @property
    def flat_list(self):
        return self._flat_list

    @property
    def confirmed_vulnerabilities(self):
        return self._confirmed_vulnerabilities

    @confirmed_vulnerabilities.setter
    def confirmed_vulnerabilities(self, confirmed_vulnerabilities: List[Vulnerability]):
        """
        Updates the vuln finding results for `PythonDependencies`
        """
        self._confirmed_vulnerabilities = confirmed_vulnerabilities

    @property
    def policy_violations(self):
        return self._policy_violations

    @policy_violations.setter
    def policy_violations(self, policy_violations=[]):
        self._policy_violations = policy_violations

    def to_json(self):
        """
        Dumps object to json.
        """
        return json.dumps(self.__dict__, default=complex_handler)


def complex_handler(obj):
    """
    Handles nested serialization to JSON.
    """
    if hasattr(obj, "to_json"):
        return obj.to_json()
    raise TypeError(
        "Object of type {} with value of {} is not JSON serializable".format(
            type(obj), repr(obj)
        )
    )



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class AffectedVersion:
    version_value: str
    operator: str


@dataclass
class Impact:
    cvss2_score: Optional[float]
    cvss2_severity: Optional[str]
    vector_string: Optional[str]
    av: Optional[str]
    ac: Optional[str]
    pr: Optional[str]
    c: Optional[str]
    i: Optional[str]
    a: Optional[str]
    ui: Optional[str]
    s: Optional[str]
    impact_score: Optional[float]
    exploitability_score: Optional[float]
    cvss3_score: Optional[float]
    cvss3_severity: Optional[str]


@dataclass
class Vulnerability:
    name: str
    cve_id: str
    cwe_id: str
    publish_date: str
    owner: str
    latest_version: str
    language: str
    description: str
    license: str
    repository_summary: str
    repo_url: str
    ochrona_severity_score: float
    found_version: str
    reason: Optional[str]
    affected_versions: List[AffectedVersion]
    references: List[str]
    impact: Impact



from typing import Any, Dict, List

from ochrona.const import PYTHON_LICENSE_TYPE_POLICY
from ochrona.model.dependency import Dependency
from ochrona.model.policy_violation import PolicyViolation


def evaluate(dependency_list: List[Dependency], policy: Dict[str, Any]):
    """
    Evaluate python license dependencies
    """
    violations = []
    allowed = [p.strip() for p in policy.get("allow_list", "").split(",") if p] or []
    deny = [p.strip() for p in policy.get("deny_list", "").split(",") if p] or []
    if len(allowed) > 0:
        # Test using allow-list approach
        for dep in dependency_list:
            if dep.license_type not in allowed:
                violations.append(
                    PolicyViolation(
                        policy_type=PYTHON_LICENSE_TYPE_POLICY,
                        friendly_policy_type="Python License Type",
                        message=f"'{dep.license_type}' not in list of allowed licenses. (from {dep.full})",
                    )
                )
    else:
        # Test using deny-list approach
        for d in deny:
            for dep in dependency_list:
                if d == dep.license_type:
                    violations.append(
                        PolicyViolation(
                            policy_type=PYTHON_LICENSE_TYPE_POLICY,
                            friendly_policy_type="Python License Type",
                            message=f"'{d}' is a restricted license type based on policy. (from {dep.full})",
                        )
                    )
    return violations


SCHEMA = {"name": PYTHON_LICENSE_TYPE_POLICY, "fields": ["allow_list", "deny_list"]}



import datetime
import re
from typing import List

from packaging.specifiers import SpecifierSet, Version

from ochrona.model.confirmed_vulnerability import Vulnerability
from ochrona.utils import parse_version_requirements

VERSION_PATTERN = r"^([a-z_]+)([\<\=|\>\=|\=\=|\=|\<|\>|\!\=]*)([\d\.]*)$"


def evaluate(vulns, required_packages) -> List[Vulnerability]:
    """
    Evalutes the required packages and potentially matching vulnerabilities
    and returns a list of confirmed vulnerabilities (if existing).

    :param vulns: Vulns found that match the required packages.
    :param required_packages: The required dependencies to evaluate against.
    :return: dict
    """
    try:
        results = []
        simple_required_packages = parse_version_requirements(required_packages)
        if simple_required_packages:
            for package_name, package_details in simple_required_packages.items():
                if not any(v.get("name") == package_name for v in vulns):
                    continue
                for vulnerability in [
                    v for v in vulns if v.get("name") == package_name
                ]:
                    # Populate vulnerable version lookups
                    vuln_specifiers = SpecifierSet()
                    exact_vuln_values = set(
                        [
                            v["version_value"]
                            for v in vulnerability["affected_versions"]
                            if v["operator"] == "="
                        ]
                    )
                    for versions in vulnerability["affected_versions"]:
                        if versions["operator"] == "=":
                            vuln_specifiers &= f"=={versions['version_value']}"
                        else:
                            vuln_specifiers &= (
                                f"{versions['operator']}{versions['version_value']}"
                            )
                    # Find latest version and requirement operator
                    dependency_version = (
                        package_details["version"]
                        if "version" in package_details
                        else ""
                    )

                    # Evaluate
                    if "-" in [s.version for s in vuln_specifiers]:  # type: ignore
                        if package_details:
                            vulnerability["found_version"] = (
                                package_name
                                + package_details.get("operator", "")
                                + package_details.get("version", "")
                            )
                        else:
                            vulnerability["found_version"] = package_name
                        vulnerability[
                            "reason"
                        ] = f"Flagged as a confirmed vulnerability because {package_name} is a required dependency and it has no known remediated versions."
                        results.append(Vulnerability(**vulnerability))
                    elif (
                        Version(dependency_version) in vuln_specifiers
                        or dependency_version in exact_vuln_values
                    ):
                        vulnerability["found_version"] = (
                            package_name
                            + package_details.get("operator", "")
                            + package_details.get("version", "")
                        )
                        vulnerability[
                            "reason"
                        ] = f"Flagged as a confirmed vulnerability because version was an exact match for dependency: {package_name}"
                        results.append(Vulnerability(**vulnerability))
        return results
    except Exception as ex:
        raise Exception("evauluate exception") from ex



# -*- coding: utf-8 -*-
"""
Ochrona-cli
:author: ascott
"""

PIPFILE_LOCK = "Pipfile.lock"
POETRY_LOCK = "poetry.lock"
SETUP_PY = "setup.py"
REQUIREMENTS_TXT = "requirements.txt"
CONSTRAINTS_TXT = "constraints.txt"
CONDA_ENVIRONMENT = "environment.yml"
TOX_INI = "tox.ini"

SUPPORTED_DEPENDENCY_FILE_PATTERNS = {
    REQUIREMENTS_TXT: "**/*requirements*.txt",
    CONSTRAINTS_TXT: "**/*constraints*.txt",
    PIPFILE_LOCK: "**/*Pipfile.lock",
    POETRY_LOCK: "**/*poetry.lock",
    SETUP_PY: "**/*setup.py",
    CONDA_ENVIRONMENT: "**/environment.yml",
    TOX_INI: "**/tox.ini",
}

INVALID_REQUIREMENTS_LINES = [
    "#",
    "-i",
    "-f",
    "-Z",
    "--index-url",
    "--extra-index-url",
    "--find-links",
    "--no-index",
    "--allow-external",
    "--allow-unverified",
    "--always-unzip",
    "--hash",
]

TOX_LINKED_REQUIREMENTS = "-r"
INVALID_TOX_LINES = ["{"]

INVALID_SPECIFIERS = {"<", ">", "!", "~"}
EQUALS_SPECIFIER = "=="

LICENSE_MAP = {
    "AGPL-3.0-only": ["AGPLv3", "AGPL 3.0"],
    "Apache-2.0": [
        "Apache License 2.0",
        "Apache License",
        "Apache 2.0",
        "Apache 2",
        "Apache Software License",
        "Apache License Version 2.0",
        "Apache License, Version 2.0",
        "Apache",
        "Apache Software License 2.0",
        "http://www.apache.org/licenses/LICENSE-2.0",
        "ASL 2",
        "APL 2.0",
        "BSD License (Apache)",
    ],
    "ZPL-2.1": ["ZPL 2.1", "ZPL"],
    "MIT": ["MIT", "MIT license", "MIT License", "MIT/Expat"],
    "HPND": ["HPND"],
    "Beerware": ["Custom BSD Beerware"],
    "BSD-2-Clause": ["BSD 2-Clause License"],
    "BSD-3-Clause": [
        "BSD-3-Clause",
        "3 Clause BSD",
        "New BSD",
        "BSD 3-Clause License",
        "3-Clause BSD License",
        "3 Clause BSD",
        "BSD3",
        "BSD License",
        "BSD",
    ],
    "LGPL-2.0-only": ["LGPLv2"],
    "LGPL-2.1-only": ["LGPL-2.1", "LGPLv2 and GPLv2+"],
    "LGPL-2.1-or-later": ["LGPL-2.1+", "LGPLv2+"],
    "LGPL-3.0-only": ["LGPL3"],
    "LGPL-3.0-or-later": ["LGPLv3+"],
    "GPL-1.0-only": ["GPL"],
    "GPL-2.0-only": ["GPL version 2", "GNU GPLv2", "GPLv2"],
    "GPL-2.0-or-later": [
        "GPLv2-or-later with a special exception which allows to use PyInstaller to build and distribute non-free programs (including commercial ones)",
        "GNU GPLv2 or any later version",
        "GPLv2+",
        "GNU GPL",
    ],
    "GPL-3.0-or-later": ["GPL-3+", "GPLv3+", "GNU/GPL version 3"],
    "GPL-3.0-only": [
        "GPL-3.0",
        "GNU General Public License v3.0",
        "GNUv3",
        "GNU GPL v3.0",
        "GPL-3",
        "GPLv3",
    ],
    "PSF-2.0": ["PSF", "PSF license"],
    "PostgreSQL": ["PostgreSQL"],
    "ISC": ["ISC License"],
    "MPL-2.0": ["MPL"],
    "Zlib": ["Zlib"],
    "Other": [
        "Other",
        "Public domain",
        "OSI Approved",
        "Copyright (C) 2008-2019 by Vinay Sajip. All Rights Reserved. See LICENSE.txt for license.",
        "BSD-derived",
        "BSD-derived (http://www.repoze.org/LICENSE.txt)",
    ],
    "EFL-2.0": ["Eiffel Forum License, version 2"],
    "BSD-3-Clause OR Apache-2.0": ["BSD or Apache License, Version 2.0"],
    "MPL-2.0 OR MIT": ["MPLv2.0, MIT Licences"],
    "Unknown": [""],
}

PYTHON_PACKAGE_NAME_POLICY = "package_name"
PYTHON_LICENSE_TYPE_POLICY = "license_type"



import dateutil.parser
import re

from packaging.specifiers import Version
from packaging.version import parse
from typing import Any, Dict, Tuple

from ochrona.client import pypi_fetch
from ochrona.const import LICENSE_MAP

PEP_SUPPORTED_OPERATORS = r"==|>=|<=|!=|~=|<|>"


class Dependency:
    """
    A python dependency object.
    """

    _raw: str = ""
    _name: str = ""
    _version: str = ""
    _version_major: str = ""
    _version_minor: str = ""
    _version_release: str = ""
    _operator: str = ""
    _full: str = ""
    _latest_version: str = ""
    _license_type: str = ""
    _latest_update: str = ""
    _release_count: str = ""
    _is_reference: bool = False

    def __init__(self, dependency: str):
        self._raw = dependency.split(",")[0]
        parts = re.split(PEP_SUPPORTED_OPERATORS, self._raw)
        if len(parts) == 1:
            if "txt" in parts[0] and "-r" in parts[0]:
                self.is_reference = True
                return
            self._name = parts[0]
        elif len(parts) > 1:
            self._name = parts[0]
            self._parse_version(parts[1])
            self._operator = re.sub("[a-zA-Z0-9.-]", "", self._raw)  # TODO fix
        self._full = self._provided_or_most_recent() or self._raw
        (
            self.latest_version,
            self.license_type,
            self.latest_update,
            self.release_count,
        ) = self._pypi_details()

    def _parse_version(self, version: str):
        v = Version(version)
        self._version = v.base_version
        version_parts = v.release
        if len(version_parts) == 1:
            self._version_major = str(version_parts[0])
        elif len(version_parts) == 2:
            self._version_major = str(version_parts[0])
            self._version_minor = str(version_parts[1])
        else:
            self._version_major = str(version_parts[0])
            self._version_minor = str(version_parts[1])
            self._version_release = str(version_parts[2])

    def _pypi_details(self) -> Tuple[str, str, str, str]:
        """
        Calls to pypi to resolve transitive dependencies.
        """
        json_value = pypi_fetch(self._name)
        if json_value:
            latest_version = self._parse_latest_version(json_value)
            license_value = self._get_license(json_value)
            latest_release_date = self._parse_latest_update(json_value, latest_version)
            release_count = self._parse_release_count(json_value)
            return latest_version, license_value, latest_release_date, release_count
        return "", "Unknown", "", ""

    def to_json(self) -> Dict[str, Any]:
        """
        Returns PythonDependency as a dict so it can be json deserialized.
        """
        return self.__dict__

    def _parse_latest_version(self, resp: Dict[str, Any]) -> str:
        version = resp.get("info", {}).get("version")
        if version is not None:
            return version
        # Fallback and check releases
        releases = resp.get("releases", {})
        if len(releases) > 1:
            return list(releases.keys())[-1]
        return ""

    def _parse_latest_update(self, resp: Dict[str, Any], latest_version: str) -> str:
        """
        Return when the latest version was published
        """
        latest_release = resp.get("releases", {}).get(latest_version)
        if latest_release is not None and isinstance(latest_release, list):
            release_artifact_dates = []
            for artifact in latest_release:
                try:
                    upload_time = artifact.get("upload_time_iso_8601")
                    parsed_upload_time = dateutil.parser.isoparse(upload_time)
                    release_artifact_dates.append(parsed_upload_time)
                except Exception:
                    pass
            latest_artifact_timestamp = max(release_artifact_dates)
            return latest_artifact_timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")
        return ""

    def _parse_release_count(self, resp: Dict[str, Any]) -> str:
        """
        Returns the total number of releases
        """
        return f"{len(resp.get('releases', []))}"

    def _get_license(self, resp: Dict[str, Any]) -> str:
        raw_license = resp.get("info", {}).get("license", None)
        if raw_license is not None:
            for ltype, matches in LICENSE_MAP.items():
                if raw_license in matches:
                    return ltype
            return raw_license
        return "Unknown"

    def _provided_or_most_recent(self) -> str:
        """
        During resolution we should decide if the provided dependency, or the
        most recent version of the dependency should be returned.

        ex.
            Required if pkg>=1.0.1
            Current version is pkg==1.2.4
            We should return pkg==1.2.4
        """

        if self._operator == ">=" and parse(self._version) <= parse(
            self._latest_version
        ):
            return f"{self._name}=={self._latest_version}"
        elif (
            self._operator == "" and self._version == "" and self._latest_version != ""
        ):
            return f"{self._name}=={self._latest_version}"
        return self._raw

    @property
    def full(self) -> str:
        return self._full

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version



from typing import Dict, List


def parse_version_requirements(packages: str) -> Dict[str, Dict[str, str]]:
    """
    Parses a list of requirements in the format <package_name><PEP_OPERATOR><SEM_VER>
    into a dictionary with the format.

    {
        <package_name>: {
            "operator": <PEP_OPERATOR>,
            "version": <SEM_VER>
        }
    }
    :param packages: List of packages with operators and semver
    :return: dict
    """

    def _parse_by_operator(operator: str, package: str) -> List[str]:
        return package.split(operator)

    parsed_results = {}
    potential_operators = ["!=", "==", ">=", "<=", ">", "<", "="]
    for package in packages:
        found = False
        for oper in potential_operators:
            if oper in package and not found:
                parsed = _parse_by_operator(oper, package)
                parsed_results[parsed[0]] = {"operator": oper, "version": parsed[1]}
                found = True
        if not found:
            parsed_results[package] = {}
    return parsed_results



from ochrona.const import PYTHON_PACKAGE_NAME_POLICY
from ochrona.model.policy_violation import PolicyViolation
from ochrona.utils import parse_version_requirements


def evaluate(dependency_list, policy):
    """
    Evaluates python package name dependencies
    """

    violations = []
    allowed = [p.strip() for p in policy.get("allow_list", "").split(",") if p] or []
    deny = [p.strip() for p in policy.get("deny_list", "").split(",") if p] or []
    simple_required_packages = parse_version_requirements(dependency_list)

    if len(allowed) > 0:
        # Test using allow-list approach
        for pkg in list(simple_required_packages.keys()):
            if pkg not in allowed:
                violations.append(
                    PolicyViolation(
                        policy_type=PYTHON_PACKAGE_NAME_POLICY,
                        friendly_policy_type="Python Package Name",
                        message=f"'{pkg}' not in list of allowed packages. (from {pkg}{simple_required_packages[pkg].get('operator', '')}{simple_required_packages[pkg].get('version', '')})",
                    )
                )
    else:
        # Test using deny-list approach
        for d in deny:
            if d in list(simple_required_packages.keys()):
                violations.append(
                    PolicyViolation(
                        policy_type=PYTHON_PACKAGE_NAME_POLICY,
                        friendly_policy_type="Python Package Name",
                        message=f"'{d}' is a restricted package based on policy. (from {d}{simple_required_packages[d].get('operator', '')}{simple_required_packages[d].get('version', '')})",
                    )
                )
    return violations


SCHEMA = {"name": PYTHON_PACKAGE_NAME_POLICY, "fields": ["allow_list", "deny_list"]}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from setuptools import setup
from ochrona import __author__ as author
from ochrona import __email__ as email
from ochrona import __license__ as license_
from ochrona import __version__ as version


try:
    with open("README.md", encoding="utf8") as readme_file:
        readme = readme_file.read()
except TypeError:
    with open("README.md") as readme_file:
        readme = readme_file.read()

requirements = [
    "click>=7.1.2",
    "pyyaml>=5.4.1",
    "requests>=2.25.0",
    "toml>=0.10.2",
    "python-dateutil>=2.8.1",
    "tarsafe>=0.0.3",
    "packaging>=20.4",
    "requests-cache>=0.5.2",
    "appdirs>=1.4.4",
]

test_requirements = ["pytest>=6.1.2"]

setup(
    name="ochrona",
    version=version,
    description="Ochrona checks your open source dependencies for vulnerabilities ",
    long_description=readme,
    long_description_content_type="text/markdown",
    author=author,
    author_email=email,
    url="https://github.com/ochronasec/ochrona-cli",
    packages=[
        "ochrona",
        "ochrona.cli",
        "ochrona.client",
        "ochrona.config",
        "ochrona.db",
        "ochrona.eval",
        "ochrona.eval.policy",
        "ochrona.eval.vuln",
        "ochrona.file",
        "ochrona.importer",
        "ochrona.log",
        "ochrona.model",
        "ochrona.parser",
        "ochrona.reporter",
        "ochrona.utils",
    ],
    package_dir={"ochrona": "ochrona"},
    entry_points={"console_scripts": ["ochrona=ochrona.cli:run"]},
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=requirements,
    license=license_,
    zip_safe=False,
    keywords="ochrona, security, dependencies, vulnerability, testing, sca",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    tests_require=test_requirements,
)



import json
import re

from ochrona.db import VulnDB
from ochrona.eval.vuln import evaluate
from ochrona.eval.policy import policy_evaluate
from ochrona.log import OchronaLogger
from ochrona.model.dependency_set import DependencySet
from ochrona.model.dependency import Dependency

PEP_SUPPORTED_OPERATORS: str = r"==|>=|<=|!=|~=|<|>"
PEP_SUPPORTED_OPERATORS_CAPTURE: str = r"(==|>=|<=|!=|~=|<|>)"


def resolve(logger: OchronaLogger, dependencies=[], policies=[]) -> DependencySet:
    """
    Resolves a list of dependencies into a PythonDependencies object
    """
    db = VulnDB(logger=logger)
    resp = DependencySet([Dependency(dep) for dep in dependencies])
    vulns = []
    for dep in resp.flat_list:
        vulns += db.lookup_by_name(_safe_query_name(dep))

    if vulns:
        resp.confirmed_vulnerabilities = evaluate(vulns, resp.flat_list)

    if len(policies) > 0:
        resp.policy_violations = policy_evaluate(dependencies=resp, policies=policies)
    return resp


def _safe_query_name(package_name):
    return re.split(PEP_SUPPORTED_OPERATORS, package_name)[0]


def _dedupe_dependencies(flat_list):
    """
    Deduplicates the flat list of dependencies to a true set resolving to what we believe is the actual resolution.
    """
    dependency_dict = collections.OrderedDict()
    for dependency in flat_list:
        simple_name = re.split(PEP_SUPPORTED_OPERATORS, dependency)[0]
        if simple_name not in dependency_dict:
            dependency_dict[simple_name] = [dependency]
        else:
            dependency_dict[simple_name].append(dependency)

    final_list = []
    for package_name, version_list in dependency_dict.items():
        final_list.append(_return_best_fit(package_name, *version_list))
    return final_list


def _return_best_fit(base, *versions):
    """
    Returns the best fit semver.

    Note: entries should be somewhat in order of appearance.
    """
    # if only one entry exists, return that
    if len(versions) == 1:
        return versions[0]
    # filter out entries with no version details
    filtered = list(
        filter(lambda v: v if re.search(PEP_SUPPORTED_OPERATORS, v) else None, versions)
    )
    # if all entries were lacking version details, return the base package name
    if not filtered:
        return base
    filtered_split = list(
        map(lambda v: re.split(PEP_SUPPORTED_OPERATORS_CAPTURE, v), filtered)
    )
    operator_list = list(map(lambda o: o[1], filtered_split))
    if "==" in operator_list:
        # If we need an exact match we should take the highest, not the first
        exact_versions = list(
            sorted(
                [
                    [parse(pkg[2]), i]
                    for i, pkg in enumerate(filtered_split)
                    if pkg[1] == "=="
                ],
                key=lambda x: x[0],
                reverse=True,
            )
        )
        return filtered[exact_versions[0][1]]
    if "<=" in operator_list:
        # If we have a less than situation we need to determine the upper bound
        upper_bound = filtered_split.index("<=")[2]
    # Finally return the highest in the list that does not violate the upper bound
    gt_versions = list(
        sorted(
            [
                [parse(pkg[2]), i]
                for i, pkg in enumerate(filtered_split)
                if pkg[1] == ">="
            ],
            key=lambda x: x[0],
            reverse=True,
        )
    )
    if len(gt_versions) > 0:
        return filtered[gt_versions[0][1]]
    return filtered[0]
